/******************************************************************************* 
 * 
 * Copyright 2011 Zack Grossbart 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ******************************************************************************/
package com.grossbart.jslim;

import java.io.IOException;
import java.io.InputStream;

import com.google.javascript.jscomp.CompilationLevel;

import org.apache.commons.io.IOUtils;
import org.junit.Test;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

/**
 * This test class covers the full unit tests for JSlim.
 */
public class JSlimTest
{

    /**
     * The basic compile test does a simple compile.
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void basicCompileTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("basic.js", readFile("basic.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("func1", funcs[0]);
    }
    
    /**
     * The recursion test makes sure we remove functions if the only references 
     * to them are recursive .
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void recursionTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("recursion.js", readFile("recursion.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("func1", funcs[0]);
    }
    
    /**
     * The recursion chain test makes sure we remove functions if the only references 
     * to them are recursive even if they are declared in a function chain. 
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void recursionChainTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("recursionchain.js", readFile("recursionchain.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("func1", funcs[0]);
    }
    
    /**
     * The recursion chain call test makes sure we remove functions if the only references 
     * to them are recursive even if they are declared in a function chain and that we 
     * don't remove functions in that case if they're called from somewhere else. 
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void recursionChainCallTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("recursionchaincall.js", readFile("recursionchaincall.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(3, funcs.length);
        assertEquals("func1", funcs[0]);
        assertEquals("func3", funcs[1]);
    }
    
    /**
     * Tests for functional property assignment.
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void funcPropsTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("propfuncs.js", readFile("propfuncs.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("func1", funcs[0]);
    }
    
    /**
     * Tests for functions which return closures.
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void funcClosureTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("closurereturn.js", readFile("closurereturn.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(2, funcs.length);
        assertEquals("func1_1", funcs[0]);
        assertEquals("func1", funcs[1]);
    }
    
    /**
     * Test function assignments.
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void assignmentTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("assignment.js", readFile("assignment.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("func2", funcs[0]);
    }
    
    /**
     * Test function arguments.
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void argumentTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("argument.js", readFile("argument.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("func2", funcs[0]);
    }
    
    /**
     * Test functions by reference.
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void referenceTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("reference.js", readFile("reference.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("func2", funcs[0]);
    }
    
    /**
     * Test functions by reference chain
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void referenceChainTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("referencechain.js", readFile("referencechain.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("func2", funcs[0]);
    }
    
    /**
     * Test functions as returned objects.
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void functionReturnTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        
        /*
         This test references a function as a property and we can't follow
         that.  This is a place where we need an extern reference.
         */
        slim.addExtern("func2");
        
        String out = slim.addLib("functionreturn.js", readFile("functionreturn.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(2, funcs.length);
        assertEquals("func1", funcs[1]);
        assertEquals("func2", funcs[0]);
    }
    
    /**
     * Check the other type of functions as returned objects.
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void functionReturn2Test()
        throws IOException
    {
        JSlim slim = new JSlim();
        
        String out = slim.addLib("functionreturn2.js", readFile("functionreturn2.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("func1", funcs[0]);
        
        /*
         The internally returned function in this test isn't "interesting" to
         our parser so it doesn't show up in the keepers list, but we still
         want to make sure that it is in the output.
         */
        assertTrue("The func2 function should still be in the output", out.indexOf("func2") > -1);
    }
    
    /**
     * Test inline functions.
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void inlineFunctionReturnTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("inlinefunctionreturn.js", readFile("inlinefunctionreturn.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("func1", funcs[0]);
        
        /*
         Now we need to make sure the newf function is still there
         */
        assertTrue("The newf function should still be in the output", out.indexOf("newf") > -1);
    }
    
    /**
     * Test functions assigned to a reference array.
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void functionRefArrayTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("functionrefarray.js", readFile("functionrefarray.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("func1", funcs[0]);
    }
    
    /**
     * Test a property assignment chain.
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void propertyAssignmentChainTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("propertyassignmentchain.js", readFile("propertyassignmentchain.js"), 
                                 CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(3, funcs.length);
        assertEquals("func3", funcs[0]);
        
        /*
         Now we need to make sure the func1 function is still there
         */
        assertTrue("The func1 function should still be in the output", out.indexOf("func1") > -1);
    }
    
    /**
     * Test a property reference by string index
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void stringPropertyReferenceTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("stringpropertyreference.js", readFile("stringpropertyreference.js"), 
                                 CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("hello", funcs[0]);
    }
    
    /**
     * Test a property reference by concatenated string index
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void stringPropertyCatReferenceTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("stringpropertycatreference.js", readFile("stringpropertycatreference.js"), 
                                 CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("hello", funcs[0]);
    }
    
    /**
     * Test a property reference by concatenated string index with a number
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void stringPropertyNumCatReferenceTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("stringpropertynumcatreference.js", readFile("stringpropertynumcatreference.js"), 
                                 CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("hello1", funcs[0]);
    }
    
    /**
     * Test a function assignment by property reference
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void functionArrayAssignmentTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("functionarrayassignment.js", readFile("functionarrayassignment.js"), 
                                 CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(2, funcs.length);
        assertEquals("func2", funcs[0]);
        assertEquals("func3", funcs[1]);
    }
    
    /**
     * Test a function assignment by property reference with a variable
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void functionArrayAssignmentVarTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("functionarrayassignmentvar.js", readFile("functionarrayassignmentvar.js"), 
                                 CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(2, funcs.length);
        assertEquals("func2", funcs[0]);
        assertEquals("func3", funcs[1]);
    }
    
    /**
     * The call tests makes sure we can track down the call method
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void callTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String out = slim.addLib("call.js", readFile("call.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        String funcs[] = slim.getKeptFunctions();
        
        assertEquals(1, funcs.length);
        assertEquals("func1", funcs[0]);
    }
    
    /**
     * Check a file with a parse error.
     * 
     * @exception IOException if there is any error reading the sample file
     */
    @Test
    public void parseErrorTest()
        throws IOException
    {
        JSlim slim = new JSlim();
        String output = slim.addLib("invalid.js", readFile("invalid.js"), CompilationLevel.SIMPLE_OPTIMIZATIONS);
        
        //slim.getErrorManager().generateReport();
        
        assertEquals(1, slim.getErrorManager().getErrorCount());
        assertEquals("JSC_TRAILING_COMMA", slim.getErrorManager().getErrors()[0].getType().key);
    }
    
    private String readFile(String name)
        throws IOException
    {
        InputStream in = getClass().getClassLoader().getResourceAsStream(name);
        try {
            return IOUtils.toString(in, "UTF-8");
        } finally {
            if (in != null) {
                in.close();
            }
        }
    }
}
